"""
This module contains the implementation of the
FR Test Statistic based estimate and the
FNN based estimate for the Bayes Error Rate

Learning to Bound the Multi-class Bayes Error (Th. 3 and Th. 4)
https://arxiv.org/abs/1811.06419
"""

from abc import abstractmethod
from typing import Tuple

import numpy as np
from scipy.sparse import coo_matrix

from daml._internal.metrics.aria.base import _BaseMetric
from daml._internal.metrics.outputs import BEROutput

from .utils import compute_neighbors, get_classes_counts, minimum_spanning_tree


class _MultiClassBer(_BaseMetric):
    def __init__(self, data: np.ndarray, labels: np.ndarray) -> None:
        self.data = data
        self.labels = labels

    @abstractmethod
    def _multiclass_ber(self, X: np.ndarray, y: np.ndarray) -> Tuple[float, float]:
        """Abstract method for the implementation of multiclass BER calculation"""

    def evaluate(self) -> BEROutput:
        """
        Calculates the Bayes Error Rate estimate

        Returns
        -------
        BEROutput
            The estimated upper and lower bounds of the Bayes Error Rate

        Raises
        ------
        ValueError
            If unique classes M < 2
        """
        # Pass X through an autoencoder before evaluating BER
        ber, ber_lower = self._multiclass_ber(self.data, self.labels)
        return BEROutput(ber=ber, ber_lower=ber_lower)


class MultiClassBerMST(_MultiClassBer):
    """
    An FR test statistic based estimator for Multi-class Bayes Error Rate

    Parameters
    ----------
    data : np.ndarray
        Array of images or image embeddings
    labels : np.ndarray
        Array of labels for each image or image embedding

    See Also
    --------
    `Learning to Bound the Multi-class Bayes Error (Th. 3 and Th. 4) <https://arxiv.org/abs/1811.06419>`_
    """  # noqa F401

    def _multiclass_ber(self, X: np.ndarray, y: np.ndarray) -> Tuple[float, float]:
        M, N = get_classes_counts(y)

        tree = coo_matrix(minimum_spanning_tree(X))
        matches = np.sum([y[tree.row[i]] != y[tree.col[i]] for i in range(N - 1)])
        deltas = matches / (2 * N)
        upper = 2 * deltas
        lower = ((M - 1) / (M)) * (1 - (1 - 2 * ((M) / (M - 1)) * deltas) ** 0.5)
        return upper, lower


class MultiClassBerFNN(_MultiClassBer):
    """
    A KNN test statistic based estimator for Multi-class Bayes Error Rate

    Parameters
    ----------
    data : np.ndarray
        Array of images or image embeddings
    labels : np.ndarray
        Array of labels for each image or image embedding

    See Also
    --------
    `Learning to Bound the Multi-class Bayes Error (Th. 3 and Th. 4) <https://arxiv.org/abs/1811.06419>`_
    """  # noqa F401

    def _multiclass_ber(self, X: np.ndarray, y: np.ndarray) -> Tuple[float, float]:
        M, N = get_classes_counts(y)

        # All features belong on second dimension
        X = X.reshape((X.shape[0], -1))
        nn_indices = compute_neighbors(X, X)
        deltas = float(np.count_nonzero(y[nn_indices] - y) / (2 * N))
        upper = 2 * deltas
        lower = ((M - 1) / (M)) * (1 - (1 - 2 * ((M) / (M - 1)) * deltas) ** 0.5)
        return upper, lower



"""
This module contains the implementation of the
FR Test Statistic based estimate for the upperbound
average precision using empirical mean precision (UAP_EMP)
"""

import numpy as np
from sklearn.metrics import average_precision_score

from daml._internal.metrics.aria.base import _BaseMetric
from daml._internal.metrics.outputs import UAPOutput


class UAP_EMP(_BaseMetric):
    """
    FR Test Statistic based estimate of the empirical mean precision

    Parameters
    ----------
    labels : np.ndarray
        A numpy array of n_samples of class labels with M unique classes.

    scores : np.ndarray
        A 2D array of class probabilities per image
    """

    def __init__(self, labels: np.ndarray, scores: np.ndarray) -> None:
        self.labels = labels
        self.scores = scores

    def evaluate(self) -> UAPOutput:
        """
        Returns
        -------
        UAPOutput
            The empirical mean precision estimate

        Raises
        ------
        ValueError
            If unique classes M < 2
        """
        uap = float(
            average_precision_score(self.labels, self.scores, average="weighted")
        )
        return UAPOutput(uap=uap)



"""
This module contains the implementation of Dp Divergence
using the First Nearest Neighbor and Minimum Spanning Tree algorithms
"""

from abc import abstractmethod

import numpy as np

from daml._internal.metrics.aria.base import _BaseMetric
from daml._internal.metrics.outputs import DivergenceOutput

from .utils import compute_neighbors, minimum_spanning_tree


class _DpDivergence(_BaseMetric):
    def __init__(self, data_a: np.ndarray, data_b: np.ndarray) -> None:
        self.data_a = data_a
        self.data_b = data_b

    @abstractmethod
    def _calculate_errors(self, data: np.ndarray, labels: np.ndarray) -> int:
        """Abstract method for the implementation of divergence calculation"""

    def evaluate(self) -> DivergenceOutput:
        """
        Calculates the divergence and any errors between the datasets

        Returns
        -------
        DivergenceOutput
            Dataclass containing the dp divergence and errors during calculation
        """
        N = self.data_a.shape[0]
        M = self.data_b.shape[0]

        stacked_data = np.vstack((self.data_a, self.data_b))
        labels = np.vstack([np.zeros([N, 1]), np.ones([M, 1])])

        errors = self._calculate_errors(stacked_data, labels)
        dp = max(0.0, 1 - ((M + N) / (2 * M * N)) * errors)
        return DivergenceOutput(dpdivergence=dp, error=errors)


class DpDivergenceMST(_DpDivergence):
    """
    A minimum spanning tree implementation of dp divergence

    Parameters
    ----------
    data_a : np.ndarray
        Array of images or image embeddings to compare
    data_b : np.ndarray
        Array of images or image embeddings to compare

    See Also
    --------
        For more information about this divergence, its formal definition,
        and its associated estimators see https://arxiv.org/abs/1412.6534.

    Warning
    -------
        MST is very slow in this implementation, this is unlike matlab where
        they have comparable speeds
        Overall, MST takes ~25x LONGER!!
        Source of slowdown:
        conversion to and from CSR format adds ~10% of the time diff between
        1nn and scipy mst function the remaining 90%
    """

    # TODO
    # - validate the input algorithm
    # - improve speed for MST, requires a fast mst implementation
    # mst is at least 10x slower than knn approach

    def _calculate_errors(self, data: np.ndarray, labels: np.ndarray) -> int:
        """
        Returns the divergence between two datasets using a minimum spanning tree

        Returns
        -------
        int
            Number of edges connecting the two datasets

        Note
        ----
            We add a small constant to the distance matrix to ensure scipy interprets
            the input graph as fully-connected.
        """
        mst = minimum_spanning_tree(data).toarray()
        edgelist = np.transpose(np.nonzero(mst))
        errors = np.sum(labels[edgelist[:, 0]] != labels[edgelist[:, 1]])
        return errors


class DpDivergenceFNN(_DpDivergence):
    """
    A first nearest neighbor implementation of dp divergence

    Parameters
    ----------
    data_a : np.ndarray
        Array of images or image embeddings to compare
    data_b : np.ndarray
        Array of images or image embeddings to compare

    See Also
    --------
        For more information about this divergence, its formal definition,
        and its associated estimators see https://arxiv.org/abs/1412.6534.
    """

    def _calculate_errors(self, data: np.ndarray, labels: np.ndarray) -> int:
        """
        Returns the divergence between two datasets using first nearest neighbor

        Parameters
        ----------
        data : np.ndarray
            Array containing images from two datasets
        labels : np.ndarray
            Array showing which dataset each image belonged to

        Returns
        -------
        int
            Number of edges connecting the two datasets
        """
        nn_indices = compute_neighbors(data, data)
        errors = np.sum(np.abs(labels[nn_indices] - labels))
        return errors
